/**
 * @ingroup   file68_devel
 * @file      SC68file.c
 * @author    Ben(jamin) Gerard <ben@sashipa.com>
 * @date      1998/09/03
 * @brief     sc68 disk file access
 * @version   $Id: SC68file.c,v 1.1.1.1 2001/12/24 20:50:10 ben Exp $
 */

/*
 *                         sc68 - Disk file access
 *         Copyright (C) 2001 Ben(jamin) Gerard <ben@sashipa.com>
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "file68/SC68file.h"
#include "file68/SC68error.h"
#include "file68/SC68utils.h"

/* SC68 file identifier string
 */
char SC68file_idstr[] = SC68_IDSTR;

/* Peek Big Endian Unaligned 32 bit value */
static int LPeek(void *a)
{
  int r;
  unsigned char *c = (unsigned char *) a;
  r = c[0] + (c[1] << 8) + (c[2] << 16) + (c[3] << 24);
  return r;
}

/* Poke Big Endian Unaligned 32 bit value */
static void LPoke(void *a, int r)
{
  unsigned char *c = (unsigned char *) a;
  c[0] = r;
  c[1] = r >> 8;
  c[2] = r >> 16;
  c[3] = r >> 24;
}

/* Verify header , return # byte to alloc & read
 */
static int read_header(FILE * f)
{
  char id[256];
  int l = sizeof(SC68file_idstr), l2;
  /* Check ID string */
  if (SC68_fread(id, (int) l, f) || SC68utils_strcmp(id, SC68file_idstr))
    return SC68error_add("Not SC68 file : Missing ID");

  /* Check 1st chunk */
  if (SC68_fread(id, 4, f) || memcmp(id, CH68_CHUNK CH68_BASE, 4))
    return SC68error_add("Not SC68 file : Missing BASE Chunk");
  if (SC68_fread(id, 4, f) || (l2 = LPeek(id), l2 <= 8))
    return SC68error_add("Not SC68 file : Weird BASE Chunk size");

  return l2 - 8;
}

static char noname[] = SC68_NOFILENAME;

/* This function init all pointers for this music files
 * It setup non initialized data to defaut value.
 */
static int valid(disk68_t * mb)
{
  music68_t *m;
  int i, previousdatasz = 0;
  void *previousdata = NULL;
  char *author = noname;
  char *composer = NULL;
  char *mname = mb->name;

  if (mb->nb_six <= 0)
    return SC68error_add("No music defined");

  /* Ensure default music in valid range */
  if (mb->default_six < 0 || mb->default_six >= mb->nb_six)
    mb->default_six = 0;

  /* No name : set default */
  if (mb->name == NULL)
    mb->name = noname;

  /* Disk total time : 00:00 */
  mb->time = 0;

  /* Clear flags */
  mb->flags = 0;

  /* Init all music in this file */
  for (m = mb->mus, i = 0; m < mb->mus + mb->nb_six; m++, i++) {
    /* default load address */
    if (m->a0 == 0)
      m->a0 = SC68_LOADADDR;
    /* default replay frequency is 50Hz */
    if (m->frq == 0)
      m->frq = 50;
    /* default time is unvalid */
    if (m->time == 0)
      m->time = -1;
    else
      mb->time += m->time;
    /* default mode is YM2149 (Atari ST) */
    if (*(int *) (&m->flags) == 0)
      m->flags.ym = 1;
    mb->flags |= *(int *) (&m->flags);
    /* default music name is file name */
    mname = m->name = (m->name == NULL) ? mname : m->name;
    /* default author */
    author = m->author = (m->author == NULL) ? author : m->author;
    /* default composer is author */
    composer = m->composer = (m->composer == NULL) ?
	((composer == NULL) ? m->author : composer) : m->composer;
    /* use data from previuous music */
    if (m->data == NULL) {
      m->data = (char *) previousdata;
      m->datasz = previousdatasz;
    }
    if (m->data == NULL)
      return SC68error_add("music #%d as no data", i);
    previousdata = m->data;
    previousdatasz = m->datasz;
  }
  /* No time : Set as unvalid */
  if (mb->time == 0)
    mb->time = -1;;

  return 0;
}


/* Check if file is probable SC68 file
 * return 0:SC68-file
 */
int SC68file_verify(char *name)
{
  FILE *f;
  int len;
  if (f = SC68_fopen(name, "rb"), f == NULL)
    return -1;
  len = read_header(f);
  SC68_fclose(f);
  return len < 0 ? len : 0;
}

/* Load , allocate memory and valid struct for SC68 music
 */
disk68_t *SC68file_load(char *name)
{
  disk68_t *mb = NULL;
  FILE *f = NULL;
  int len, room;
  int chk_size;
  music68_t *cursix;
  char *b;

  if (f = SC68_fopen(name, "rb"), f == NULL)
    goto error;
  if (len = read_header(f), len < 0)
    goto error;

  room = len + sizeof(disk68_t);
  if (mb = (disk68_t *) SC68_malloc(room), mb == NULL)
    goto error;
  memset(mb, 0, room);
  mb->data = (char *) (mb + 1);
  if (SC68_fread(mb->data, (int) len, f))
    goto error;

  for (b = mb->data, cursix = 0; len >= 8; b += chk_size, len -= chk_size) {
    char chk[8];
    if (b[0] != 'S' || b[1] != 'C')
      break;
    chk[0] = b[2];
    chk[1] = b[3];
    chk[2] = 0;
    chk_size = LPeek(b + 4);
    b += 8;
    len -= 8;

    if (!strcmp(chk, CH68_BASE)) {
    }
    /* Music general info */
    else if (!strcmp(chk, CH68_DEFAULT)) {
      mb->default_six = LPeek(b);
    } else if (!strcmp(chk, CH68_FNAME)) {
      mb->name = b;
    }

    /* start music session */
    else if (!strcmp(chk, CH68_MUSIC)) {
      /* More than 256 musix !!! : Prematured end */
      if (mb->nb_six >= 256) {
	len = 0;
	break;
      }
      cursix = mb->mus + mb->nb_six;
      mb->nb_six++;
    }
    /* Music name */
    else if (!strcmp(chk, CH68_MNAME)) {
      if (cursix == NULL)
	goto error;
      cursix->name = b;
    }
    /* Author name */
    else if (!strcmp(chk, CH68_ANAME)) {
      if (cursix == NULL)
	goto error;
      cursix->author = b;
    }
    /* Composer name */
    else if (!strcmp(chk, CH68_CNAME)) {
      if (cursix == NULL)
	goto error;
      cursix->composer = b;
    }
    /* External replay */
    else if (!strcmp(chk, CH68_REPLAY)) {
      if (cursix == NULL)
	goto error;
      cursix->replay = b;
    }
    /* 68000 D0 init value */
    else if (!strcmp(chk, CH68_D0)) {
      if (cursix == NULL)
	goto error;
      cursix->d0 = LPeek(b);
    }
    /* 68000 memory load address */
    else if (!strcmp(chk, CH68_AT)) {
      if (cursix == NULL)
	goto error;
      cursix->a0 = LPeek(b);
    }
    /* Playing time */
    else if (!strcmp(chk, CH68_TIME)) {
      if (cursix == NULL)
	goto error;
      cursix->time = LPeek(b);
    }
    /* Replay frequency */
    else if (!strcmp(chk, CH68_FRQ)) {
      if (cursix == NULL)
	goto error;
      cursix->frq = LPeek(b);
    }
    /* replay flags */
    else if (!strcmp(chk, CH68_TYP)) {
      if (cursix == NULL)
	goto error;
      *(int *) &cursix->flags = LPeek(b);
    }
    /* music data */
    else if (!strcmp(chk, CH68_MDATA)) {
      if (cursix == NULL)
	goto error;
      cursix->data = b;
      cursix->datasz = chk_size;
    }
    /* EOF */
    else if (!strcmp(chk, CH68_EOF)) {
      len = 0;
      break;
    }
  }

  /* Check it */
  if (len) {
    SC68error_add("%s : Corrupted SC68 file", name);
    goto error;
  }
  if (valid(mb))
    goto error;
  return mb;

error:
  SC68_fclose(f);
  SC68_free(mb);
  return NULL;
}

/* save CHUNK & data */
static int save_chunk(FILE * f, char *chunk, void *data, int size)
{
  int err;
  chunk68_t chk;
  memcpy(chk.id, CH68_CHUNK, 2);
  memcpy(chk.id + 2, chunk, 2);
  LPoke(chk.size, size);
  if (err = SC68_fwrite(&chk, (int) sizeof(chunk68_t), f), err < 0)
    return err;
  if (size > 0 && data != NULL)
    err = SC68_fwrite(data, size, f);
  return err;
}

/* save CHUNK & string (only if non-NULL & lenght>0) */
static int save_string(FILE * f, char *chunk, char *str)
{
  int len;
  if (str == NULL)
    return 0;
  len = strlen(str);
  if (len == 0)
    return 0;
  return save_chunk(f, chunk, str, len + 1);
}

/* save CHUNK & string str ( only if oldstr!=str & lenght>0 ) */
static int save_differstr(FILE * f, char *chunk, char *str, char *oldstr)
{
  int len;
  if (oldstr == str)
    return 0;
  if (str == NULL)
    return 0;
  if (oldstr != NULL && !strcmp(oldstr, str))
    return 0;
  len = strlen(str);
  if (len == 0)
    return 0;
  return save_chunk(f, chunk, str, len + 1);
}

/* save CHUNK and 4 bytes Big Endian integer */
static int save_number(FILE * f, char *chunk, int n)
{
  char number[4];
  LPoke(number, n);
  return save_chunk(f, chunk, number, 4);
}

/* save CHUNK and number ( only if n!=0) */
static int save_nonzero(FILE * f, char *chunk, int n)
{
  if (!n)
    return 0;
  return save_number(f, chunk, n);
}

/* Save disk into file
 */
int SC68file_save(char *fname, disk68_t * mb)
{
  FILE *f = NULL;
  int err = 0;
  music68_t *mus;
  long pos, len;
  char *mname, *aname, *cname, *data;

  if (fname == NULL || mb == NULL)
    return SC68error_add("SC68file_save(%p,%p) NULL pointer(s)", fname,
			 mb);
  if (mb->nb_six <= 0 || mb->nb_six > SC68_MAX_TRACK)
    return SC68error_add("SC68file_save(%s) : Bad number of track (%d)",
			 fname, mb->nb_six);
  if (f = SC68_fopen(fname, "wb"), f == NULL)
    return SC68error_add("SC68file_save(%s) Failed", fname);

  /* SC68 file header string */
  if (err =
      SC68_fwrite(SC68file_idstr, (int) sizeof(SC68file_idstr), f),
      err < 0)
    goto error;
  /* Save pos for further calculation */
  pos = (long) ftell(f);
  /* SC68 1st chunk */
  if (err = save_chunk(f, CH68_BASE, NULL, 0), err < 0)
    goto error;
  /* Disk name */
  if (err = save_string(f, CH68_FNAME, mb->name), err < 0)
    goto error;
  /* Default track */
  if (err = save_nonzero(f, CH68_DEFAULT, mb->default_six), err < 0)
    goto error;

  /* Reset previous value for variuos string */
  mname = mb->name;
  aname = cname = data = NULL;
  for (mus = mb->mus; mus < mb->mus + mb->nb_six; mus++) {
    /* New track */
    if (err = save_chunk(f, CH68_MUSIC, NULL, 0), err < 0)
      goto error;
    /* Name, author... */
    if (err = save_differstr(f, CH68_MNAME, mus->name, mname), err < 0)
      goto error;
    if (err = save_differstr(f, CH68_ANAME, mus->author, aname), err < 0)
      goto error;
    if (err = save_differstr(f, CH68_CNAME, mus->composer, cname), err < 0)
      goto error;
    if (err = save_string(f, CH68_REPLAY, mus->replay), err < 0)
      goto error;
    if (mus->name != NULL)
      mname = mus->name;
    if (mus->author != NULL)
      aname = mus->author;
    if (mus->composer != NULL)
      cname = mus->composer;

    /* Play parms */
    if (err = save_nonzero(f, CH68_D0, mus->d0), err < 0)
      goto error;
    if (err =
	save_nonzero(f, CH68_AT, mus->a0 == SC68_LOADADDR ? 0 : mus->a0),
	err < 0)
      goto error;
    if (err = save_nonzero(f, CH68_TIME, mus->time), err < 0)
      goto error;
    if (err =
	save_nonzero(f, CH68_FRQ, mus->frq == 50 ? 0 : mus->frq), err < 0)
      goto error;
    if (err = save_number(f, CH68_TYP, *(int *) &mus->flags), err < 0)
      goto error;

    /* Music Data */
    if (mus->data != data && mus->data != NULL) {
      if (err = save_chunk(f, CH68_MDATA, mus->data, mus->datasz), err < 0)
	goto error;
      data = mus->data;
    }
  }
  /* SC68 last chunk */
  if (err = save_chunk(f, CH68_EOF, NULL, 0), err < 0)
    goto error;
  /* Write total lenght */
  len = ftell(f) - pos;
  fseek(f, pos, SEEK_SET);
  if (err = save_chunk(f, CH68_BASE, NULL, (int) len), err < 0)
    goto error;

  SC68_fclose(f);
  return 0;

error:
  SC68_fclose(f);
  return err;
}
