/**
 * @ingroup   io68_devel
 * @file      paula_io.c
 * @author    Ben(jamin) Gerard <ben@sashipa.com>
 * @date      1998/07/18
 * @brief     Paula IO plugin (AMIGA soundchip)
 * @version   $Id: paula_io.c,v 1.1.1.1 2001/12/24 20:50:10 ben Exp $
*/

/*
 *                sc68 - Paula IO plugin (AMIGA soundchip)
 *         Copyright (C) 1999 Ben(jamin) Gerard <ben@sashipa.com>
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "io68/paula_io.h"
#include "io68/paulaemul.h"

/* About Amiga hardware registers
  -------------------------------

 Amiga hw register could have different address for
 read and write access. It is the case for DMA control
 register.
 Register could be byte or word wide. Anyway I assume
 it is written in memory order !!

 SET/CLR rules:
 -------------
 bit S (MSB) is SET/CLR bit.
 if S is SET (=1), set all setted bit (OR)
 else clear all setted bit (NAND)

DFF096 | S0000000 \
DFF097 | 0000DCBA / DMA audio write

DFF002 | 00000000 \
DFF003 | 0000DCBA / DMA audio read

DFF09A -> S0000DCB \
DFF09B -> A0000000 / DMA IRQ ( don't care )

VOICE:
	+0 = sample start
	+4 = sample len ( in word )
	+6 = Period  		( 12 bit )
	+8 = Volume  		( 0-64 )

*/

static void audio_int(unsigned int intreq);
static void reload(paulav_t *v , u8 *p);

static unsigned int clearset(unsigned int v, unsigned int clrset)
{
  if(clrset&0x8000)
  	return v | (clrset&0x7FFF);
  else
  	return v & ~clrset;
}

static unsigned int DMACON(int dmacon)
{
	return (dmacon << (8*sizeof(int)-1-9) >> (8*sizeof(int)-1) ) & dmacon & 0xF;
}

/*
static unsigned int INTREQ(void)
{
	return paula_intreq&(0xF<<7);
}
*/

static unsigned int INTENA(int intena)
{
	return (intena << (8*sizeof(int)-1-14) >> (8*sizeof(int)-1)) & (intena) & (0xF<<7);
}

static u32 paula_readB(u32 addr, u32 cycle)
{
	addr=addr; cycle=cycle;
  addr&=0xFF;
	switch(addr&0xFF)
	{
		case PAULA_VHPOSR:
//			audio_int( 0x8000 | ((DMACON(paula_dmacon)^15)<<7) );
			return paula[PAULA_VHPOSR]++;

		case PAULA_DMACONRH: 	return (paula_dmacon>>8)&0x7f;
		case PAULA_DMACONRL:  return paula_dmacon&0xFF;
		case PAULA_INTENARH:	return (paula_intena>>8)&0x7f;
		case PAULA_INTENARL:	return paula_intena&0xFF;
		case PAULA_INTREQRH:	return (paula_intreq>>8)&0x7f;
		case PAULA_INTREQRL:	return paula_intreq&0xFF;
		default:							return paula[addr];
	}
}

static u32 paula_readW(u32 addr, u32 cycle)
{
	return (paula_readB(addr,cycle)<<8) | paula_readB(addr+1,cycle);
}

static u32 paula_readL(u32 addr, u32 cycle)
{
	return (paula_readW(addr,cycle)<<16) | paula_readW(addr+2,cycle+4);
}

/* start DMA on voice n ( DMA disable -> enable ) */
static void start_DMA(int bit)
{
	int old,chg,cur;
	old = DMACON(paula_dmacon);
	paula_dmacon |= bit;
	cur = DMACON(paula_dmacon);
	chg = cur & ~old;

	if(chg&1) reload(paulav+0, paula+0xA0);
	if(chg&2) reload(paulav+1, paula+0xB0);
	if(chg&4) reload(paulav+2, paula+0xC0);
	if(chg&8) reload(paulav+3, paula+0xD0);
/*	printf("START(%1X,%1X)\n",old,cur);
*/
}

/* stop DMA bit to be stopped */
static void stop_DMA( int bit )
{
	paula_dmacon &= ~bit;
//	audio_int(0x8000 | ((bit^0xF)<<7));

/*	printf("STOP(%X)\n",bit&0x20f);
*/

}

/* Reload paula internal register with current value */
static void reload(paulav_t *v , u8 *p)
{
	unsigned len;
	v->start =
	v->adr   = (((p[1]<<16)|(p[2]<<8)|p[3])&0x7FFFE)<<PAULA_CT_FIX;
	len      = (p[4]<<(8) | p[5]) << (PAULA_CT_FIX+1);
	len      = (len==0) ? (1<<(17+PAULA_CT_FIX)) : len;
	v->end   = v->start + len;
/*	printf("reload %02X: %x %d\n",p-paula, v->adr>>PAULA_CT_FIX, len>>PAULA_CT_FIX );
*/
}

/* Check wheter sound interrupt is denied, which
 * case paula internal counter must be reload
*/
static void audio_int(unsigned int intreq)
{
	unsigned int intdenied;
	if(!(intreq&0x8000))
	{
		paula_intreq &= ~intreq;
		return;
	}
	/* Master interrupt not set : DENIED */
	intdenied = ~INTENA(paula_intena);
	/* Already requested : DENIED */
	intdenied |= paula_intreq;
	/* Only interrested by requested int */
	intdenied &= intreq;
	if(intdenied & (1<< 7)) reload(paulav+0, paula+0xA0);
	if(intdenied & (1<< 8)) reload(paulav+1, paula+0xB0);
	if(intdenied & (1<< 9)) reload(paulav+2, paula+0xC0);
	if(intdenied & (1<<10)) reload(paulav+3, paula+0xD0);
	paula_intreq |= intreq;
}

static void paula_writeB(u32 addr, u32 v, u32 cycle)
{
	addr=addr; cycle=cycle;
	addr &= 0xFF;
	v    &= 0xFF;
	paula[addr] = v;

	switch(addr)
	{
		case PAULA_DMACONL:
		{
			int i;

			i = (paula[PAULA_DMACONH]<<8) | v;
			if(i&0x8000) start_DMA(i&0x20F);
			else stop_DMA(i&0x20F);
/*			printf("DMACON : %X\n", paula_dmacon);
*/
			break;
		}

		/* INTENA */
		case PAULA_INTENAL:
			paula_intena = clearset(paula_intena,(paula[PAULA_INTENAH]<<8)|v);
			break;

		/* INTREQ */
		case PAULA_INTREQL:
			audio_int( (paula[PAULA_INTREQH]<<8)|v );
/*			printf("INTREQ : %X (%X)\n",INTREQ(),(paula[PAULA_INTREQH]<<8)|v);
*/
			break;

		default:
			break;
	}
}

static void paula_writeW(u32 addr, u32 v, u32 cycle)
{
	paula_writeB(addr,   v>>8, cycle);
	paula_writeB(addr+1, v,    cycle);
}

static void paula_writeL(u32 addr, u32 v, u32 cycle)
{
	paula_writeW(addr,   v>>16, cycle);
	paula_writeW(addr+2, v,     cycle+4);
}

static int68_t *paula_int( u32 cycle )
{
	cycle = cycle;
	return 0L;
}

static u32 paula_nextint( u32 cycle )
{
	cycle = cycle;
	return IO68_NO_INT;
}

static u32 paula_reset(void)
{
  paula_dmacon=0x200;
	PL_reset();
	return 0;
}

static void paula_subcycle( u32 sub )
{
	sub = sub;
}

io68_t paula_io =
{
	NULL,
		"AMIGA Paula",
		0xFFDFF000, 0xFFDFF0DF,
		{paula_readB,  paula_readW,  paula_readL},
		{paula_writeB, paula_writeW, paula_writeL},
		paula_int, paula_nextint,
		paula_subcycle,
		paula_reset,
		0,0
};
